import Test.QuickCheck
import Data.Char

instance Arbitrary Char where
    arbitrary     = choose ('\32', '\128')
    coarbitrary c = variant (ord c `rem` 4)

-- Definition der Cursorposition
-- Anm: '|' stellt den Cursor dar
-- "|abc" = (0,"abc") = ("","abc")
-- "a|bc" = (1,"abc") = ("a","bc")
-- "ab|c" = (2,"abc") = ("ba","c")
-- "abc|" = (3,"abc") = ("cba","")

type Buffer = (Int,String)

-- the empty buffer
empty 	:: Buffer
empty = (0, "")

-- insert character before cursor
insert :: Char -> Buffer -> Buffer
insert c (cursor,string) = (cursor+1,left ++ [c] ++ right)
	where
	(left,right) = splitAt cursor string

-- delete character before cursor
delete :: Buffer -> Buffer
delete (0, string)      = (0, string)
delete (cursor, string) = (cursor-1, left ++ right)
	where
	left = take (cursor-1) string
	right = drop cursor string

-- move cursor left one character
left :: Buffer -> Buffer
left (0, string)      = (0,string)
left (cursor, string) = (cursor-1,string)

-- move cursor right one character
right :: Buffer -> Buffer
right (cursor, string)
  | cursor == (length string) = (cursor,string)
  | otherwise                 = (cursor+1,string)

-- is cursor at left end?
atLeft :: Buffer -> Bool
atLeft (cursor, _) = cursor == 0

-- is cursor at right end?
atRight :: Buffer -> Bool
atRight (cursor, string) = cursor == length string

-- #######################################################

type BufferI = (String,String)

emptyI :: BufferI
emptyI = ("","")

insertI :: Char -> BufferI -> BufferI
insertI c (s,t) = (c:s, t)

deleteI :: BufferI -> BufferI
deleteI ([],t)   = ([],t)
deleteI (x:xs,t) = (xs,t)

rightI :: BufferI -> BufferI
rightI (s,[])   = (s,[])
rightI (s,x:xs) = (x:s, xs)

leftI :: BufferI -> BufferI
leftI ([],t)   = ([],t)
leftI (x:xs,t) = (xs, x:t)

atLeftI :: BufferI -> Bool
atLeftI (s,_) = length s == 0

atRightI :: BufferI -> Bool
atRightI (_,t) = length t == 0

-- #######################################################

retrieve :: BufferI -> Buffer
retrieve (left,right) = (length left, reverse left ++ right)

-- #######################################################

prop_empty = retrieve emptyI == empty

prop_atLeft :: BufferI -> Bool
prop_atLeft q = atLeftI q == atLeft (retrieve q)

prop_atRight :: BufferI -> Bool
prop_atRight q = atRightI q == atRight (retrieve q)

prop_insert :: Char -> BufferI -> Bool
prop_insert c b = retrieve (insertI c b) == insert c (retrieve b)

prop_delete :: BufferI -> Bool
prop_delete b = retrieve (deleteI b) == delete (retrieve b)

prop_right :: BufferI -> Bool
prop_right b = retrieve (rightI b) == right (retrieve b)

prop_left :: BufferI -> Bool
prop_left b = retrieve (leftI b) == left (retrieve b)
