-- Computes the stream of powers of 2
pof2s :: [Integer]
pof2s = 1 : map (*2) pof2s

pof2test :: [Integer]
pof2test = 1 : [x*2 | x <- pof2test ]

-- Computes the stream of rows of the pascal's triangle
pts :: [[Integer]]
pts = [1] : pts_intern [1]

-- Handles the recursion to compute the pascal's triangle
pts_intern :: [Integer] -> [[Integer]]
pts_intern l = [row] ++ pts_intern row
	where row = rowCalc l 

-- Computes the next row of the pascal's triangle
rowCalc :: [Integer] -> [Integer]
rowCalc (x:xs) = [1] ++ (zipWith (+) xs (x:xs)) ++ [1]

-- Computes the Goedel number of the input variable zahl
-- We expect that 0 is a positive number!
-- (Transform the input number into an array of its digits, zip it with the first n primes and 
-- fold the resulting array to get the result)
g :: Integer -> Integer
g zahl
	| zahl >= 0 = foldl (*) 1 $ zipWith (^) (take (length digits) primes) digits
	| otherwise = 0
	where
	digits = getDigits zahl

-- Computes the stream of prime numbers
primes :: [Integer]
primes = 2 : sieve [3,5..]
	where
	sieve (x:xs) = x : sieve [y | y <- xs, mod y x >0]

-- Transforms a number into an array of its digits
-- eg: 1234 -> [1,2,3,4]
-- 0 has to be handled seperately
getDigits :: Integer -> [Integer]
getDigits 0 = [0]
getDigits n = getDigits_ n []
	where
	getDigits_ 0 list = list
	getDigits_ n list = getDigits_ (div n 10) ((mod n 10):list)

-- Computes the stream of Goedel numbers
gs :: [Integer]
gs = map g [1..]

-- Computes all numbers which differ at most by d from their Meerten number 
dM :: Integer -> [Integer]
dM d
	| d < 0 = repeat 0
	| otherwise = dM_ d [1..]

dM_ :: Integer -> [Integer] -> [Integer]
dM_ d (x:xs)
	| abs (goedelNr - x) <= d = x : (dM_ d xs)
	| otherwise = dM_ d xs
	where
	goedelNr = g x
